/** 
 ****************************************************************************
 *
 * ServiceA service
 *
 * @file    ServiceAImpl.cpp
 *
 * @author  Charles Halliday
 *
 * @date    15 June 2010
 *
 ****************************************************************************
 */

#include "ServiceAImpl.hpp"

#include "com/sun/star/bridge/XUnoUrlResolver.hpp"

#include <iostream>

#define OUSTRING_TO_ASCII(oustring) \
     OUStringToOString((oustring), \
                        RTL_TEXTENCODING_ASCII_US).getStr()

#define PRINT_ERROR(message) { \
    cerr << "ERROR: uno_client: " \
         << message                \
         << endl ;                 \
}

#define HANDLE_UNO_EXC(ex, txt_for_system_log) { \
        const std::string type  = \
        OUSTRING_TO_ASCII(getCppuType(&ex).getTypeName()); \
        PRINT_ERROR("HANDLED "                                  \
                  << type << ": " << txt_for_system_log << std::endl \
                  << type << ".Message = " << OUSTRING_TO_ASCII(ex.Message) \
                  ); \
        }

using namespace ::com::sun::star::lang;
using namespace ::com::sun::star::uno;
using namespace ::rtl;
using namespace ::std;

namespace {
  using namespace ::com::sun::star::bridge;
  using namespace ::test;

  
  const OUString sHostname =  OUString::createFromAscii("127.0.0.1");
  const OUString sServiceBName = OUString::createFromAscii("test.ServiceB");
  const sal_Int32 sServiceBPort = 8102;
  const OUString sServiceCName = OUString::createFromAscii("test.ServiceC");
  const sal_Int32 sServiceCPort = 8101;


Reference <XInterface> getRemoteServiceInstance(
  const Reference< XComponentContext > & localContext,
  const ::rtl::OUString & hostname,
  const sal_Int32  port,
  const ::rtl::OUString & serviceName)
  throw (::com::sun::star::uno::Exception)
{
  Reference <XInterface> remoteObject(0);

  Reference < XMultiComponentFactory > localServiceManager(0);
  OUString unoUrl =
      OUString::createFromAscii("uno:")
      + OUString::createFromAscii("socket,host=") + hostname
      + OUString::createFromAscii(",port=")
      + OUString::valueOf((sal_Int32)((sal_uInt16)port)) // no negative values
      + OUString::createFromAscii(",tcpNoDelay=1")
      + OUString::createFromAscii(";urp,Negotiate=0,ForceSynchronous=0;")
      + serviceName;

  const OUString urlResolverServiceName =
    OUString::createFromAscii("com.sun.star.bridge.UnoUrlResolver");

  localServiceManager = localContext->getServiceManager();
  assert(localServiceManager.is());

  Reference < XInterface > urlResolverXInterface =
    localServiceManager->createInstanceWithContext(urlResolverServiceName,
						   localContext);

  assert(urlResolverXInterface.is());
    
  try {
    Reference < XUnoUrlResolver > urlResolver(urlResolverXInterface,
					      UNO_QUERY );
    assert(urlResolver.is());

    remoteObject = urlResolver->resolve(unoUrl);
    assert(remoteObject.is());

  }
  catch ( ::com::sun::star::connection::ConnectionSetupException &uex ) {
    HANDLE_UNO_EXC(uex,
		   "Couldn't access local resource url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( IllegalArgumentException &uex ) {
    HANDLE_UNO_EXC(uex,
		   "Uno url invalid, url="
		   <<  OUStringToOString(unoUrl,
					 RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( ::com::sun::star::connection::NoConnectException & uex ) {
    HANDLE_UNO_EXC(uex,
		   "No server listening on the resource, url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( RuntimeException & uex ) {
    HANDLE_UNO_EXC(uex,
		   "A remote call was aborted, url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }

  return remoteObject;
} // ! getRemoteServiceInstance

}

namespace test {

  ServiceAImpl::ServiceAImpl(
    Reference< XComponentContext > const & xComponentContext)
    : ServiceAImplBase::ServiceAImplBase(xComponentContext)
    , mXComponentContext(xComponentContext)
  {
  }

  ServiceAImpl::~ServiceAImpl()
  {
  }

  // Interface com.sun.star.lang.XInitialization

  // Method of com.sun.star.lang.XInitialization
  void SAL_CALL
  ServiceAImpl::initialize(
    const Sequence< Any >& aArguments )
    throw (
      Exception,
      RuntimeException )
  {
  }

  // Interface com.sun.star.lang.XMain

  // Method of com.sun.star.lang.XMain
  sal_Int32 SAL_CALL
  ServiceAImpl::run(
    const Sequence< OUString >& aArguments )
    throw ( RuntimeException )
  {
    sal_Int32 returnValue = 0;
    cerr << __PRETTY_FUNCTION__ << " entered" << endl;

    // Call ServiceB XMain::run
    try {
      Reference <XInterface> xInterface =
	getRemoteServiceInstance(mXComponentContext,
				 sHostname,
				 sServiceBPort,
				 sServiceBName);
      Reference <XMain> serviceBxMain ( xInterface, UNO_QUERY);
      if ( serviceBxMain.is() ) {
	Sequence < OUString > runArgs(0);
	serviceBxMain->run(runArgs);
      }
      else {
	PRINT_ERROR("Could not get ServiceB");
      }
      // explicit clear
      xInterface.clear();
      serviceBxMain.clear();
    }
    catch (::com::sun::star::uno::Exception & ex) {
      HANDLE_UNO_EXC(ex,"Error in ServiceB");
    }
    catch ( ... ) {
      PRINT_ERROR("Unknown exception in ServiceB");
    }


     // Call ServiceC XMain::run
    try {
      Reference <XInterface> xInterface =
	getRemoteServiceInstance(mXComponentContext,
				 sHostname,
				 sServiceCPort,
				 sServiceCName);
      Reference <XMain> serviceCxMain ( xInterface, UNO_QUERY);
      if ( serviceCxMain.is() ) {
	Sequence < OUString > runArgs(0);
	serviceCxMain->run(runArgs);
      }
      else {
	PRINT_ERROR("Could not get ServiceC");
      }
      // explicit clear
      xInterface.clear();
      serviceCxMain.clear();
    }
    catch (::com::sun::star::uno::Exception & ex) {
      HANDLE_UNO_EXC(ex,"Error in ServiceC");
    }
    catch ( ... ) {
      PRINT_ERROR("Unknown exception in ServiceC");
    }
   

    return returnValue;
  }

  // Additional class member implementations


} // namespace close

