/**
 * Test application to 
 * Run:
 *   If -run option given in command line, then calls run() with
 *   the arguments in -runargs option.
 */

#include <unistd.h>
#include <netdb.h>
#include <iostream>
#include <string.h>
#include <stdlib.h>

#include <rtl/ustring.hxx>
#include <rtl/strbuf.hxx>
#include <assert.h>

#include <osl/diagnose.h>
#include <osl/file.h>

#include <cppuhelper/bootstrap.hxx>
#include <cppuhelper/servicefactory.hxx>

// generated c++ interfaces
#include "com/sun/star/bridge/XUnoUrlResolver.hpp"
#include <com/sun/star/lang/XComponent.hpp>
#include <com/sun/star/uno/XComponentContext.hpp>
#include <com/sun/star/lang/XMain.hpp>
#include <com/sun/star/lang/XMultiComponentFactory.hpp>

#include "com/sun/star/connection/ConnectionSetupException.hpp"
#include "com/sun/star/connection/NoConnectException.hpp"
#include "com/sun/star/lang/IllegalArgumentException.hpp"
#include "com/sun/star/uno/RuntimeException.hpp"

using namespace com::sun::star::bridge;
using namespace com::sun::star::uno;
using namespace com::sun::star::lang;

using namespace ::rtl;
using namespace ::std;


namespace
{

// Option names
#define OPT_UNORC     "-unorc"
#define OPT_VERBOSITY "-verbosity"
#define OPT_PORT      "-port"
#define OPT_HOST      "-host"
#define OPT_SERVICE   "-service"

#define UNO_URL_SCHEMA    "uno:"

#define OUSTRING_TO_ASCII(oustring) \
     OUStringToOString((oustring), \
                        RTL_TEXTENCODING_ASCII_US).getStr()

#define PRINT_ERROR(message) { \
    cerr << "ERROR: uno_client: " \
         << message                \
         << endl ;                 \
}

#define HANDLE_UNO_EXC(ex, txt_for_system_log) { \
        const std::string type  = \
        OUSTRING_TO_ASCII(getCppuType(&ex).getTypeName()); \
        PRINT_ERROR("HANDLED "                                  \
                  << type << ": " << txt_for_system_log << std::endl \
                  << type << ".Message = " << OUSTRING_TO_ASCII(ex.Message) \
                  ); \
        }

#define PRINT_EXC_ERROR(ex) { \
    PRINT_ERROR("Caught Exception, Message=" \
                <<  OUStringToOString(ex.Message,      \
                               RTL_TEXTENCODING_UNICODE).getStr()); \
}

// Shared variables
// Command line options
OUString  sUnorc = OUString::createFromAscii("");
sal_Int32 sVerbosity = 1;
sal_Int32 sPort = -1;
const OUString  sHost = OUString::createFromAscii("127.0.0.1");

OUString  sService = OUString::createFromAscii("");

// Local functions

void printOptions()
{
  cout << "arg values:" << endl ;
  cout << "  -unorc=" 
       << OUSTRING_TO_ASCII(sUnorc) << endl ;
  cout << "  -verbosity=" << sVerbosity << endl ;
  cout << "  -port=" << sPort << endl ;
  cout << "  -host=" 
       << OUSTRING_TO_ASCII(sHost) << endl ;
  cout << "  -service=" 
       << OUSTRING_TO_ASCII(sService) << endl ;
}

static void printUsage()
{
  cout << "usage: uno_client"
       << " [-unorc unorcfile]"
       << " [-verbosity 0..5]"
       << " -port portnumber"
       << " -service uno_service_name"
       << endl;
  cout << "-unorc specifies the uno resource file,"
       << endl ;
  cout << "-port specifies the port publishing the service" << endl;
  cout << "      if set, loads rfemote service on local host and this port"
       << endl;
  cout << "      if not set, loads service locally"
       << endl;
  cout << "-service specifies the UNO service to use, also is url object name,"
       << endl;
  cout << "calls XMain::run after loading service" << endl;
  if ( sVerbosity >= 3) {
    printOptions();
  }
}


/** Get args from command line
 */
bool getArgs(int argc, const char **argv)
{
  static const char moduleName[] = "uno_client::getArgs:" ;
  sal_Bool OK = true;

  int  nextArgNum;
  const char *nextArg;

  for ( nextArgNum = 1; OK && nextArgNum < argc; nextArgNum++) {
    nextArg =  argv[nextArgNum];
 
    if (strcmp(nextArg,OPT_UNORC) == 0 ) {
      nextArgNum++;
      if ( nextArgNum < argc) {
        sUnorc =  OUString::createFromAscii(argv[nextArgNum]);
      }
      else {
        OK = sal_False;
      }
    }
    else if (strcmp(nextArg,OPT_VERBOSITY) == 0 ) {
      nextArgNum++;
      if ( nextArgNum < argc) {
        nextArg =  argv[nextArgNum];
        char *endPtr;
        long verbosity = strtol(nextArg, &endPtr, 10);
        if ( endPtr != nextArg ) {
          sVerbosity = (sal_Int32)verbosity;
        }
        else {
          OK = sal_False;
        }
      }
      else {
        OK = sal_False;
      }
    }
    else if (strcmp(nextArg, OPT_PORT) == 0 ) {
      nextArgNum++;
      if ( nextArgNum < argc) {
        nextArg =  argv[nextArgNum];
        char *endPtr;
        long port = strtol(nextArg, &endPtr, 10);
        if ( endPtr != nextArg ) {
          sPort = (sal_Int32)port;
        }
        else {
          OK = sal_False;
        }
      }
      else {
        OK = sal_False;
      }
    }
    else if (strcmp(nextArg, OPT_SERVICE) == 0 ) {
      nextArgNum++;
      if ( nextArgNum < argc) {
        sService =  OUString::createFromAscii(argv[nextArgNum]);
      }
      else {
        OK = sal_False;
      }
    }
    else {
     // Bad/missing option
      PRINT_ERROR("Unexpected command line arg:" << nextArg);
      OK = sal_False;
    }
  }

  if ( sService.getLength() == 0 ) {
    cerr << "ERROR: "
         << moduleName
         << "Missing -service option"
         << endl;
    OK = sal_False;
  }
  if ( ! OK ) {
     printUsage();
  }
  else {
    if ( sVerbosity >= 5) {
      printOptions();
    }
  }

  cerr << "-unorc=" << OUSTRING_TO_ASCII(sUnorc)
       << " -port=" << sPort 
       << " -service=" << OUSTRING_TO_ASCII(sService)
       << endl;

  return OK;

} // ! getArgs

Reference <XInterface> getServiceInstance(
  const Reference< XComponentContext > & localContext,
  const ::rtl::OUString & serviceName)
  throw (::com::sun::star::uno::Exception)
{
  Reference <XInterface> xInterface;
  Reference < XMultiComponentFactory > localServiceManager(0);

  localServiceManager = localContext->getServiceManager();
  assert(localServiceManager.is());

  xInterface = 
    localServiceManager->createInstanceWithContext(serviceName,
						   localContext);
  assert(xInterface.is());

  return xInterface;
}

Reference <XInterface> getRemoteServiceInstance(
  const Reference< XComponentContext > & localContext,
  const ::rtl::OUString & hostname,
  const sal_Int32  port,
  const ::rtl::OUString & serviceName)
  throw (::com::sun::star::uno::Exception)
{
  Reference <XInterface> remoteObject(0);

  Reference < XMultiComponentFactory > localServiceManager(0);
  OUString unoUrl =
      OUString::createFromAscii("uno:")
      + OUString::createFromAscii("socket,host=") + hostname
      + OUString::createFromAscii(",port=")
      + OUString::valueOf((sal_Int32)((sal_uInt16)port)) // no negative values
      + OUString::createFromAscii(",tcpNoDelay=1")
      + OUString::createFromAscii(";urp,Negotiate=0,ForceSynchronous=0;")
      + serviceName;

  const OUString urlResolverServiceName =
    OUString::createFromAscii("com.sun.star.bridge.UnoUrlResolver");

  localServiceManager = localContext->getServiceManager();
  assert(localServiceManager.is());

  Reference < XInterface > urlResolverXInterface =
    localServiceManager->createInstanceWithContext(urlResolverServiceName,
						   localContext);

  assert(urlResolverXInterface.is());
    
  try {
    Reference < XUnoUrlResolver > urlResolver(urlResolverXInterface,
					      UNO_QUERY );
    assert(urlResolver.is());

    remoteObject = urlResolver->resolve(unoUrl);
    assert(remoteObject.is());

  }
  catch ( ::com::sun::star::connection::ConnectionSetupException &uex ) {
    HANDLE_UNO_EXC(uex,
		   "Couldn't access local resource url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( IllegalArgumentException &uex ) {
    HANDLE_UNO_EXC(uex,
		   "Uno url invalid, url="
		   <<  OUStringToOString(unoUrl,
					 RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( ::com::sun::star::connection::NoConnectException & uex ) {
    HANDLE_UNO_EXC(uex,
		   "No server listening on the resource, url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch ( RuntimeException & uex ) {
    HANDLE_UNO_EXC(uex,
		   "A remote call was aborted, url="
		   << OUStringToOString(unoUrl,
					RTL_TEXTENCODING_ASCII_US).getStr());
  }

  return remoteObject;
} // ! getRemoteServiceInstance


} // ! anon namespace

/**
 * main connects to remote component and reports details
 * optionally calls its XMain.run method.
 */
extern "C" SAL_CALL int main(int argc, const char **argv)
{
  static const char moduleName[] = "uno_client:" ;

  if ( ! getArgs(argc, argv) ) {
    cerr << moduleName << ": Bad arguments, exiting" << endl;
    exit(1);
  }
  int exitValue = 0;
  Reference <XInterface> serviceXInterface(0);
  Reference< XComponentContext > xComponentContext(0);

  try {
    xComponentContext =
      ::cppu::defaultBootstrap_InitialComponentContext(sUnorc);
  }
  catch (::com::sun::star::uno::Exception & ex) {
    HANDLE_UNO_EXC(ex,
		   "defaultBootstrap_InitialComponentContext failed for unorc="
		   <<  OUStringToOString(sUnorc,
					 RTL_TEXTENCODING_ASCII_US).getStr());
  }
  catch (...) {
    PRINT_ERROR("Caught unknown exception");
    exitValue = 1;
  }

  if (sPort > 0 ) {
    try {
      serviceXInterface = getRemoteServiceInstance(xComponentContext,
						  sHost,
						  sPort,
						  sService);
    }
    catch (::com::sun::star::uno::Exception & ex) {
      HANDLE_UNO_EXC(ex, "Could not get service"
		     << OUSTRING_TO_ASCII(sService)
		     << " on "
		     << OUSTRING_TO_ASCII(sHost)
		     << " port "
		     << sPort);
    }
    catch ( ... ) {
      PRINT_ERROR(
        "Instantiation of " << OUSTRING_TO_ASCII(sService)
        << " on " << OUSTRING_TO_ASCII(sHost) << " port " <<  sPort
        << " threw unknown exception");
    }
  }
  else {
    // TODO Load service locally
    serviceXInterface =  getServiceInstance(xComponentContext,
					    sService);
  }

  if (serviceXInterface.is() ) {
    // Get XMain interface and call run()
    Reference <XMain> xMain ( serviceXInterface, UNO_QUERY);
  
    if (xMain.is()) {
      cerr << moduleName << ": Calling " << OUSTRING_TO_ASCII(sService) 
	   << " XMain::run()" << endl;
      Sequence <OUString> runArgs(0);
      xMain->run(runArgs);
    }
    else {
      PRINT_ERROR("XMain not supported by "
		  << OUSTRING_TO_ASCII(sService));
    }
    xMain.clear();
    serviceXInterface.clear();
  }
  else {
    exitValue = 1;
  }

  exit(exitValue);

}  // ! main
